from django.urls import path
from enhanced_curation import views

urlpatterns = [
    path("enhanced_curation/ranked_proposals", views.ranked_proposals),
    path("enhanced_curation/csv", views.get_csv),
]



# Generated by Django 4.0.5 on 2022-06-29 23:37

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("enhanced_curation", "0005_documentcache_top25percent"),
    ]

    operations = [
        migrations.AddField(
            model_name="documentcache",
            name="has_achievement_level",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="documentcache",
            name="top1percent",
            field=models.BooleanField(default=False),
        ),
    ]



# Generated by Django 4.0.5 on 2022-06-22 18:11

import django.contrib.postgres.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("enhanced_curation", "0003_alter_collectioncache_last_run"),
    ]

    operations = [
        migrations.AddField(
            model_name="documentcache",
            name="priority_populations",
            field=django.contrib.postgres.fields.ArrayField(
                base_field=models.TextField(), default=list, size=None
            ),
        ),
        migrations.AddField(
            model_name="documentcache",
            name="subject",
            field=models.TextField(null=True),
        ),
        migrations.AddField(
            model_name="documentcache",
            name="work_locations",
            field=django.contrib.postgres.fields.ArrayField(
                base_field=models.TextField(), default=list, size=None
            ),
        ),
    ]



# Generated by Django 4.0.5 on 2022-06-22 13:36

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("enhanced_curation", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="documentcache",
            name="org_budget",
            field=models.TextField(null=True),
        ),
        migrations.AlterField(
            model_name="documentcache",
            name="org_size",
            field=models.TextField(null=True),
        ),
    ]



from django.db import models

from django.contrib.postgres.fields import ArrayField

from torque import models as torque_models
from math import floor


class CollectionCache(models.Model):
    collection = models.ForeignKey(torque_models.Collection, on_delete=models.CASCADE)
    last_run = models.DateTimeField(null=True)

    def rebuild_cache(self):
        DocumentCache.objects.filter(collection_cache=self).delete()
        cached_documents = []

        def document_score(document):
            try:
                return float(
                    document.values.get(field__name="Panel Score").to_python()[
                        "Normalized"
                    ]
                )
            except ValueError:
                return 0
            except torque_models.Value.DoesNotExist:
                pass

            try:
                return float(
                    document.values.get(field__name="Peer Score").to_python()[
                        "Normalized"
                    ]
                )
            except ValueError:
                return 0
            except torque_models.Value.DoesNotExist:
                pass

            return 0

        scores = [document_score(d) for d in self.collection.documents.all()]
        scores.sort()
        score_cutoff_25percent = scores[floor(len(scores) * 0.75)]
        score_cutoff_1percent = scores[floor(len(scores) * 0.99)]

        for document in self.collection.documents.all():
            has_achievement_level = False
            try:
                has_achievement_level = bool(
                    document.values.get(field__name="Achievement Level").to_python()
                )
            except torque_models.Value.DoesNotExist:
                pass
            try:
                cached_documents.append(
                    DocumentCache(
                        document=document,
                        collection_cache=self,
                        org_size=DocumentCache.categorize_org_size(document),
                        org_budget=DocumentCache.categorize_org_budget(document),
                        subject=DocumentCache.categorize_subject(document),
                        priority_populations=DocumentCache.categorize_priority_populations(
                            document
                        ),
                        work_locations=DocumentCache.categorize_work_locations(
                            document
                        ),
                        org_name=document.values.get(
                            field__name="Organization Name"
                        ).to_python(),
                        top25percent=(
                            score_cutoff_25percent != 0
                            and document_score(document) > score_cutoff_25percent
                        ),
                        top1percent=(
                            score_cutoff_1percent != 0
                            and document_score(document) > score_cutoff_1percent
                        ),
                        has_achievement_level=has_achievement_level,
                    )
                )
            except torque_models.Value.DoesNotExist:
                # Values here are required! (like org name)
                pass

        DocumentCache.objects.bulk_create(cached_documents)


class DocumentCache(models.Model):
    document = models.ForeignKey(torque_models.Document, on_delete=models.CASCADE)
    collection_cache = models.ForeignKey(
        CollectionCache, on_delete=models.CASCADE, related_name="documents"
    )
    org_size = models.TextField(null=True)
    org_budget = models.TextField(null=True)
    org_name = models.TextField()
    subject = models.TextField(null=True)
    priority_populations = ArrayField(models.TextField(), default=list)
    work_locations = ArrayField(models.TextField(), default=list)
    top25percent = models.BooleanField(default=False)
    top1percent = models.BooleanField(default=False)
    has_achievement_level = models.BooleanField(default=False)

    def categorize_subject(document):
        try:
            return (
                document.values.get(field__name="Primary Subject Area")
                .to_python()
                .get("Level 1", None)
            )
        except torque_models.Value.DoesNotExist:
            return None

    def categorize_priority_populations(document):
        try:
            return document.values.get(field__name="Priority Populations").to_python()
        except torque_models.Value.DoesNotExist:
            return []

    def categorize_work_locations(document):
        try:
            work_locations = []
            for location in document.values.get(
                field__name="Current Work Locations"
            ).to_python():
                region = location.get("Region", None)
                country = location.get("Country", None)
                subregion = location.get("Subregion", None)

                if (
                    region in ["Oceania", "Asia", "Europe", "Africa"]
                    and region not in work_locations
                ):
                    work_locations.append(region)

                if (
                    subregion in ["North America", "Caribbean", "Central America"]
                    and "North America" not in work_locations
                ):
                    work_locations.append("North America")

                if subregion == "South America" and subregion not in work_locations:
                    work_locations.append(subregion)

                if country == "United States" and country not in work_locations:
                    work_locations.append(country)

            return work_locations
        except torque_models.Value.DoesNotExist:
            return []

    def categorize_org_size(document):
        categories = {
            "Fewer than 10 Full-time Employees": "small",
            "10 to 25 Full-time Employees": "small",
            "26 to 50 Full-time Employees": "small",
            "51 to 100 Full-time Employees": "medium",
            "101 to 300 Full-time Employees": "medium",
            "301 to 500 Full-time Employees": "medium",
            "501 to 1,000 Full-time Employees": "large",
            "1,000+ Full-time Employees": "large",
        }

        try:
            return categories.get(
                document.values.get(field__name="Number of Employees").to_python(), None
            )
        except torque_models.Value.DoesNotExist:
            return None

    def categorize_org_budget(document):
        categories = {
            "Less than $1 Million": "small",
            "$1 to $5 Million": "small",
            "$5 to $10 Million": "small",
            "$10 to $25 Million": "small",
            "$25 to $50 Million": "medium",
            "$50 to $100 Million": "medium",
            "$100 to $500 Million": "medium",
            "$100 to $250 Million": "medium",
            "$250 to $500 Million": "medium",
            "$500 to $750 Million": "large",
            "$500 Million to $1 Billion": "large",
            "$750 Million to $1 Billion": "large",
            "$1 Billion +": "large",
        }
        try:
            return categories.get(
                document.values.get(field__name="Annual Operating Budget").to_python(),
                None,
            )
        except torque_models.Value.DoesNotExist:
            return None
        document.values.get(field__name="")

    def score(self, selection_groups):
        score = 0
        for selection_group in selection_groups:
            if selection_group["name"] == "Subject":
                for selection in selection_group["selections"]:
                    if selection == self.subject:
                        score += int(selection_group["level"])
            elif selection_group["name"] == "Populations":
                for selection in selection_group["selections"]:
                    if selection in self.priority_populations:
                        score += int(selection_group["level"])
            elif selection_group["name"] == "Work Location":
                for selection in selection_group["selections"]:
                    if selection in self.work_locations:
                        score += int(selection_group["level"])

        self.score = score

    def __eq__(self, other):
        return self.score == other.score and self.org_name == other.org_name

    def __ne__(self, other):
        return self.score != other.score or self.org_name != other.org_name

    def __gt__(self, other):
        if self.score == other.score:
            return self.org_name > other.org_name

        # This is inverted because we want higher scores to come first, but also ealier org names to come first!
        return self.score < other.score



import time
from multiprocessing import Process
from django.db import transaction
from django import db
import sys
from django.contrib.postgres.search import SearchVector
import traceback
from django.utils import timezone


class RebuildCollectionCache:
    def run(self):
        from torque import models as torque_models
        from enhanced_curation import models

        for collection in torque_models.Collection.objects.all():
            (cache, created) = models.CollectionCache.objects.get_or_create(
                collection=collection
            )

            if not cache.last_run or cache.last_run < collection.last_updated:
                # We do this, so that if it errors, we aren't looping forever
                cache.last_run = timezone.now()
                cache.save()

                print(
                    "Rebuilding Enhanced Curation Cache for colletion "
                    + collection.name
                    + "...",
                    end="",
                )
                cache.rebuild_cache()
                print("Rebuilt")


class CacheRebuilder(Process):
    def __init__(self):
        super().__init__()
        self.daemon = True

    def run(self):
        db.connections.close_all()

        while True:
            try:
                RebuildCollectionCache().run()
            except:
                print("Rebuilder failed a loop due to %s" % sys.exc_info()[0])
                print(traceback.format_exc())

            time.sleep(5)



from django.http import HttpResponse, JsonResponse, FileResponse
from torque import models as torque_models
from django.db.models import Q
from django.conf import settings
from enhanced_curation import models
import random
import json
import csv


def get_proposals(request):
    group = request.GET["group"]
    q = request.GET["q"]
    global_wiki_key = request.GET["wiki_key"]
    global_collection_name = request.GET["collection_name"]
    wiki_keys = request.GET["wiki_keys"].split(",")
    collection_names = request.GET["collection_names"].split(",")
    org_sizes = request.GET["org_sizes"].split(",") if request.GET["org_sizes"] else []
    selections = json.loads(request.GET["selections"])
    org_budgets = (
        request.GET["org_budgets"].split(",") if request.GET["org_budgets"] else []
    )
    proposal_types = (
        request.GET["proposal_types"].split(",")
        if request.GET["proposal_types"]
        else []
    )
    global_config = torque_models.WikiConfig.objects.get(
        collection__name=global_collection_name,
        wiki__wiki_key=global_wiki_key,
        group=group,
    )
    configs = torque_models.WikiConfig.objects.filter(
        collection__name__in=collection_names, wiki__wiki_key__in=wiki_keys, group=group
    ).all()

    filters = [
        Q(collection_cache__collection__name__in=collection_names),
    ]
    if "All Proposals" not in proposal_types:
        filters.append(Q(top25percent=True)),

    if len(org_sizes) > 0:
        filters.append(Q(org_size__in=org_sizes))

    if len(org_budgets) > 0:
        filters.append(Q(org_budget__in=org_budgets))

    proposal_type_q = Q()

    if "Top Proposals" in proposal_types and "All Proposals" not in proposal_types:
        proposal_type_q = proposal_type_q | Q(top1percent=True)
    if "BSN" in proposal_types and "All Proposals" not in proposal_types:
        proposal_type_q = proposal_type_q | Q(has_achievement_level=True)
    if "Non-awardees" in proposal_types and "All Proposals" not in proposal_types:
        proposal_type_q = proposal_type_q | Q(has_achievement_level=False)
    filters.append(proposal_type_q)

    if q:
        search_results = torque_models.SearchCacheDocument.objects.filter(
            collection__name__in=collection_names,
            wiki__wiki_key__in=configs.values_list("wiki__wiki_key", flat=True),
            group__in=configs.values_list("group", flat=True),
            wiki_config__in=configs,
            data_vector=q,
        ).select_related("document")

        searched_documents = [result.document for result in search_results]

        filters.append(Q(document__in=searched_documents))

    proposals = list(models.DocumentCache.objects.filter(*filters))
    for proposal in proposals:
        proposal.score(selections)

    proposals.sort()

    return proposals


def ranked_proposals(request):
    proposals = get_proposals(request)
    wiki_keys = request.GET["wiki_keys"].split(",")
    group = request.GET["group"]

    collection_names = request.GET["collection_names"].split(",")
    configs = torque_models.WikiConfig.objects.filter(
        collection__name__in=collection_names, wiki__wiki_key__in=wiki_keys, group=group
    ).all()

    resp = []
    for proposal in proposals[0:20]:
        for config in configs:
            if config.collection == proposal.document.collection:
                resp.append(proposal.document.to_dict(config))

    return JsonResponse({"result": resp, "count": len(proposals)})


def get_csv(request):
    response = HttpResponse(
        content_type="text/csv",
        headers={"Content-Disposition": 'attachment; filename="enhanced-curation.csv"'},
    )

    field_names = [
        "Competition Name",
        "Organization Name",
        "Organization Location",
        "Project Title",
        "Project Description",
        "Executive Summary",
        "Solution Overview",
        "Priority Populations",
        "Current Work Locations",
        "Future Work Locations",
        "Primary Subject Area",
        "Key Words and Phrases",
        "Sustainable Development Goals",
        "Annual Operating Budget",
        "Number of Employees",
    ]

    proposals = get_proposals(request)
    writer = csv.writer(response)

    columns = []
    for field_name in field_names:
        if field_name in getattr(settings, "TORQUE_CSV_PROCESSORS", {}):
            columns.extend(
                settings.TORQUE_CSV_PROCESSORS[field_name].field_names(field_name)
            )
        else:
            columns.append(field_name)

    writer.writerow(columns)

    for proposal in proposals:
        row = []
        values_by_field = {
            v.field.name: v for v in proposal.document.values.prefetch_related("field")
        }
        for field_name in field_names:
            if field_name in values_by_field and field_name in getattr(
                settings, "TORQUE_CSV_PROCESSORS", {}
            ):
                row.extend(
                    settings.TORQUE_CSV_PROCESSORS[field_name].process_value(
                        values_by_field[field_name].to_python()
                    )
                )
            elif field_name in getattr(settings, "TORQUE_CSV_PROCESSORS", {}):
                row.extend(
                    [""]
                    * len(
                        settings.TORQUE_CSV_PROCESSORS[field_name].field_names(
                            field_name
                        )
                    )
                )
            elif field_name in values_by_field:
                row.append(values_by_field[field_name].to_python())
            else:
                row.append("")
        writer.writerow(row)

    return response



# Generated by Django 4.0.5 on 2022-06-29 16:34

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("enhanced_curation", "0004_documentcache_priority_populations_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="documentcache",
            name="top25percent",
            field=models.BooleanField(default=False),
        ),
    ]



# Generated by Django 4.0.5 on 2022-06-22 13:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("enhanced_curation", "0002_alter_documentcache_org_budget_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="collectioncache",
            name="last_run",
            field=models.DateTimeField(null=True),
        ),
    ]



#!/usr/bin/env python

from distutils.core import setup

with open("README.md", "r", encoding="utf-8") as readme:
    long_description = readme.read()

setup(
    name="django-torque-enhanced-curation",
    version="0.2.0",
    description="django app for torque enhanced curation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Open Tech Strategies, LLC",
    author_email="frankduncan@opentechstrategies.com",  # For now, this works
    url="https://code.librehq.com/ots/mediawiki/enhanced-curation",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
    ],
    packages=[
        "enhanced_curation",
        "enhanced_curation.migrations",
        "enhanced_curation.cache_rebuilder",
    ],
    install_requires=[
        "django-torque",
    ],
    package_dir={"": "."},
    python_requres=">=3.6",
)



# Generated by Django 4.0.5 on 2022-06-22 12:44

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("torque", "0040_value_unique_value_for_field_document"),
    ]

    operations = [
        migrations.CreateModel(
            name="CollectionCache",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("last_run", models.DateTimeField()),
                (
                    "collection",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="torque.collection",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="DocumentCache",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("org_size", models.TextField()),
                ("org_budget", models.TextField()),
                ("org_name", models.TextField()),
                (
                    "collection_cache",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="documents",
                        to="enhanced_curation.collectioncache",
                    ),
                ),
                (
                    "document",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="torque.document",
                    ),
                ),
            ],
        ),
    ]



from django.apps import AppConfig
from enhanced_curation.cache_rebuilder import background
import os


class CacheRebuilderConfig(AppConfig):
    name = "enhanced_curation.cache_rebuilder"
    label = "enhanced_curation_cache_rebuilder"

    def ready(self):
        # This check makes sure that this only boots once
        if os.environ.get("RUN_MAIN", None) != "true" and not os.environ.get(
            "DISABLE_CACHE_REBUILDER", False
        ):
            cache_rebuilder = background.CacheRebuilder()
            cache_rebuilder.start()
